"""
Provides all classes for pane-based paradigms.
"""

from __future__ import annotations

import logging
import re
import string
from itertools import zip_longest
from typing import Collection, Iterable, Mapping, Optional, Sequence, TextIO

from more_itertools import ilen, one
from hfst_optimized_lookup import TransducerFile, Analysis

from paradigm_panes.settings import STRICT_GENERATOR_FST_FILEPATH, STRICT_ANALYSER_FST_FILEPATH, STRICT_GENERATOR_FST_WITH_MORPHEMES_FILEPATH

logger = logging.getLogger(__name__)


class ParseError(Exception):
    """
    Raised when there's an issue in parsing paradigm layouts.
    """


class ParadigmGenerationError(Exception):
    """
    Raised when there's some issue with generating/filling a paradigm.
    """


class ParadigmIsNotStaticError(Exception):
    """
    Raised when calling .as_static_paradigm() on a layout with unfilled
    placeholders (i.e., on a dynamic paradigm).
    """


class Paradigm:
    """
    A paradigm is intended to display a collection of wordforms related to a single
    lexeme, or a table of related wordforms.

    A paradigm in this project is composed of multiple **panes**. Each pane itself is a
    table, organized by **rows**, and then each row contains **cells**.
    """

    def __init__(self, panes: Iterable[Pane]):
        self._panes = tuple(panes)

    @property
    def panes(self) -> Iterable[Pane]:
        yield from self._panes

    @property
    def max_num_columns(self):
        """
        How many columns are necessary for this entire paradigm?
        """
        return max(pane.num_columns for pane in self.panes)

    def contains_wordform(self, wordform: str) -> bool:
        """
        True if the wordform is found ANYWHERE in the paradigm.
        """
        return any(pane.contains_wordform(wordform) for pane in self.panes)


class ParadigmLayout(Paradigm):
    """
    Layout for a particular word class. The layout contains placeholders
    (InflectionTemplate) that can be filled at runtime to generate a paradigm that
    can be rendered and shown to the user.
    """

    @property
    def inflection_cells(self) -> Iterable[InflectionTemplate]:
        for pane in self.panes:
            yield from pane.inflection_cells

    @classmethod
    def load(cls, layout_file: TextIO) -> ParadigmLayout:
        """
        Load a paradigm layout from a file.
        """
        return cls.loads(layout_file.read())

    @classmethod
    def loads(cls, text: str) -> ParadigmLayout:
        """
        Load a ParadigmLayout from a string.
        """
        lines = text.splitlines(keepends=False)

        pane_lines: list[list[str]] = [[]]
        for line in lines:
            if line.strip() == "":
                # empty line; start a new pane
                pane_lines.append([])
            else:
                # add line to current pane
                pane_lines[-1].append(line)

        pane_strs = ["\n".join(lines) for lines in pane_lines if len(lines) > 0]

        panes = [Pane.parse(p) for p in pane_strs if p.strip()]
        return ParadigmLayout(panes)

    def dumps(self):
        """
        Export the layout as a string. This string can be read by .loads().
        """

        pane_text = []
        num_columns = self.max_num_columns
        for pane in self.panes:
            pane_text.append(pane.dumps(require_num_columns=num_columns))

        tabs = "\t" * (num_columns - 1)
        empty_line = f"\n{tabs}\n"

        return empty_line.join(pane_text)

    def generate_fst_analyses(self, lemma: str) -> dict[str, str]:
        """
        Generates a dictionary mapping analysis templates to analyses substituted with
        the given lemma.
        """
        return {
            inflection.analysis_template: inflection.as_analysis(lemma)
            for inflection in self.inflection_cells
        }

    def fill(self, forms: Mapping[str, Collection[str]]) -> Paradigm:
        """
        Given a mapping from analysis to a collection of wordforms, returns a
        paradigm with all its InflectionTemplate cells replaced with WordformCells.
        """
        return Paradigm(pane.fill(forms) for pane in self.panes)

    def as_static_paradigm(self) -> Paradigm:
        """
        Returns a Paradigm that can be rendered without having to explicitly fill it.

        :raises ParadigmIsNotStaticError: when called on a dynamic paradigm
        """
        if ilen(self.inflection_cells) > 0:
            raise ParadigmIsNotStaticError(
                "not a static paradigm; contains " "inflection templates"
            )
        return self.fill({})

    def __str__(self):
        return self.dumps()


class Pane:
    """
    A self-contained table from the full paradigm.

    A pane contains a number of rows.
    """

    def __init__(self, rows: Iterable[Row]):
        self._rows = tuple(rows)

    @property
    def header(self) -> Optional[HeaderRow]:
        first_row = self._rows[0]
        if isinstance(first_row, HeaderRow):
            return first_row
        return None

    @property
    def num_columns(self) -> int:
        return max(row.num_cells for row in self.rows)

    @property
    def inflection_cells(self) -> Iterable[InflectionTemplate]:
        for row in self.rows:
            yield from row.inflection_cells

    @property
    def rows(self) -> Iterable[Row]:
        yield from self._rows

    @property
    def tr_rows(self) -> Iterable[Row]:
        """
        Yields rows needed in the HTML model. All rows of a compound row are yielded
        individually.
        """
        for row in self._rows:
            if isinstance(row, CompoundRow):
                yield from row.subrows
            else:
                yield row

    def dumps(self, require_num_columns: Optional[int] = None) -> str:
        """
        Returns a string representation that can be parsed again.
        :param require_num_columns: if given, the pane must have at least this many columns.
                                    Rows are padded with empty cells at the end.
        """
        return "\n".join(row.dumps(require_num_columns) for row in self.rows)

    def __str__(self):
        return self.dumps()

    def __eq__(self, other) -> bool:
        if isinstance(other, Pane):
            return self._rows == other._rows
        return False

    def __repr__(self):
        name = type(self).__qualname__
        rows = ", ".join(repr(row) for row in self.rows)
        return f"{name}([{rows}])"

    @classmethod
    def parse(cls, text: str) -> Pane:
        lines = text.splitlines()

        if len(lines) == 0:
            raise ParseError("Not enough lines in pane")

        return Pane(ContentRow.parse(line) for line in lines)

    def contains_wordform(self, wordform: str) -> bool:
        return any(row.contains_wordform(wordform) for row in self.rows)

    def fill(self, forms: Mapping[str, Collection[str]]) -> Pane:
        return Pane(row.fill(forms) for row in self.rows)


class Row:
    is_header: bool
    has_content: bool
    num_cells: int

    @property
    def inflection_cells(self) -> Iterable[InflectionTemplate]:
        # subclasses MUST implement this somehow
        raise NotImplementedError

    @property
    def cells(self) -> Iterable[Cell]:
        # subclasses MUST implement this somehow
        raise NotImplementedError

    def dumps(self, require_num_columns: Optional[int] = None) -> str:
        """
        Returns a string representation of the row that can be parsed again.
        :param require_num_columns: if given, the row must have at least this many columns.
                                    The row is padded with empty cells at the end.
        """

        row_as_string = str(self)
        if require_num_columns is None:
            return row_as_string

        if require_num_columns < 1:
            raise ValueError("must require at least one column")

        # Figure out how many tabs we need to add to the end.
        total_tabs = require_num_columns - 1
        tabs_present = row_as_string.count("\t")
        tabs_left = max(total_tabs - tabs_present, 0)

        return row_as_string + "\t" * tabs_left

    def contains_wordform(self, wordform: str) -> bool:
        raise NotImplementedError

    def fill(self, forms: Mapping[str, Collection[str]]) -> Row:
        if not self.has_content:
            # Just labels; can return ourselves verbatim
            return self
        # Subclass must override this (i.e., ContentRow)
        raise NotImplementedError

    @staticmethod
    def parse(text: str) -> Row:
        if text.startswith("# "):
            return HeaderRow.parse(text)

        cell_texts = text.rstrip("\n").split("\t")
        # Remove empty cells from the end.
        while cell_texts:
            if cell_texts[-1].strip() != "":
                break
            cell_texts.pop()

        if len(cell_texts) == 0:
            # This would mean the line was empty, but empty lines in a paradigm file
            # are used as pane separators.
            # If we got here from ParadigmLayout.parse(), something terrible has gone
            # wrong.
            raise ParseError("Refusing to parse empty row")

        return ContentRow(Cell.parse(t) for t in cell_texts)


class ContentRow(Row):
    """
    A single row from a pane. Rows contain cells.
    """

    is_header = False
    has_content = True

    def __init__(self, cells: Iterable[Cell]):
        self._cells = tuple(cells)

    @property
    def cells(self):
        yield from self._cells

    @property
    def num_cells(self):
        return len(self._cells)

    @property
    def inflection_cells(self) -> Iterable[InflectionTemplate]:
        return (c for c in self.cells if isinstance(c, InflectionTemplate))

    def contains_wordform(self, wordform: str) -> bool:
        return any(cell.contains_wordform(wordform) for cell in self.cells)

    def fill(self, forms: Mapping[str, Collection[str]]) -> ContentRow | CompoundRow:
        """
        Fills the column and may return a ContentRow or a CompoundRow (a composition
        of ContentRows.
        """

        # Fill each **column**, then figure out if we need to make a compound row
        columns = n_empty_lists(self.num_cells)
        for index, cell in enumerate(self.cells):
            columns[index].extend(cell.fill(forms))

        num_rows_needed = max(len(c) for c in columns)
        assert num_rows_needed > 0
        if num_rows_needed == 1:
            # A row with all columns having a single cell:
            return ContentRow(one(cells) for cells in columns)

        # Create compound rows
        rows = []
        for row_num in range(num_rows_needed):
            cells = [cell_if_exists_or_no_output(col, row_num) for col in columns]
            cells = [adjust_row_span(cell, num_rows_needed) for cell in cells]
            rows.append(ContentRow(cells))
        return CompoundRow(rows)

    def __eq__(self, other) -> bool:
        if not isinstance(other, ContentRow):
            return False
        return all(a == b for a, b in zip_longest(self.cells, other.cells))

    def __str__(self):
        try:
            return "\t".join(str(cell[0]) for cell in self.cells)
        except TypeError as e:
            return "\t".join(str(cell) for cell in self.cells)

    def __repr__(self):
        name = type(self).__qualname__
        cells_repr = ", ".join(repr(cell) for cell in self.cells)
        return f"{name}([{cells_repr}])"


class HeaderRow(Row):
    """
    A row that acts as a header for the rest of the pane.
    """

    prefix = "#"
    is_header = True
    has_content = False
    num_cells = 0

    def __init__(self, tags):
        super().__init__()
        self._tags = tuple(tags)

    @property
    def inflection_cells(self) -> Iterable[InflectionTemplate]:
        # a header, by definition, has no inflections.
        return ()

    @property
    def fst_tags(self) -> tuple[str, ...]:
        return self._tags

    def __eq__(self, other) -> bool:
        if isinstance(other, HeaderRow):
            return self._tags == other._tags
        return False

    def __str__(self):
        return " ".join(f"{self.prefix} {tag}" for tag in self._tags)

    def __repr__(self):
        name = type(self).__qualname__
        cells_repr = ", ".join(repr(cell) for cell in self._tags)
        return f"{name}([{cells_repr}])"

    def contains_wordform(self, wordform: str) -> bool:
        """
        A header **never** contains a wordform, so this always returns False.
        """
        return False

    @classmethod
    def parse(cls, text: str) -> HeaderRow:
        if not text.startswith(cls.prefix):
            raise ParseError("Not a header row: {text!r}")

        text = text.rstrip()
        return HeaderRow(parse_label(text, prefix=cls.prefix))


class CompoundRow(Row):
    """
    Contains multiple filled content rows.
    """

    is_header = False
    has_content = True

    def __init__(self, rows: Iterable[ContentRow]):
        self._rows = tuple(rows)

    @property
    def cells(self) -> Iterable[Cell]:
        for row in self._rows:
            yield from row.cells

    @property
    def num_cells(self):
        return max(row.num_cells for row in self._rows)

    @property
    def num_subrows(self) -> int:
        return len(self._rows)

    @property
    def subrows(self) -> Iterable[ContentRow]:
        return iter(self._rows)

    @property
    def inflection_cells(self) -> Iterable[InflectionTemplate]:
        raise AssertionError("cannot return inflection cells in filled row")

    def contains_wordform(self, wordform: str) -> bool:
        return any(row.contains_wordform(wordform) for row in self._rows)


class Cell:
    """
    A single cell from a paradigm.
    """

    is_label: bool = False
    is_inflection: bool = False
    is_missing: bool = False
    is_empty: bool = False
    should_suppress_output: bool = False

    @staticmethod
    def parse(text: str) -> Cell:
        if text == "":
            return EmptyCell()
        elif text == "--":
            return MissingForm()
        elif text.startswith("_ "):
            return RowLabel.parse(text)
        elif text.startswith("| "):
            return ColumnLabel.parse(text)
        elif looks_like_analysis_string(text):
            return InflectionTemplate.parse(text)
        else:
            return WordformCell.parse(text)

    def contains_wordform(self, wordform: str) -> bool:
        if not self.is_inflection:
            return False
        # Must be overridden in subclasses
        raise NotImplementedError

    def fill(self, forms: Mapping[str, Collection[str]]) -> tuple[Cell, ...]:
        """
        Returns one or more cells by filling the paradigm with the provided forms.
        """
        if not self.is_inflection:
            return (self,)
        # This should be overridden by subclasses.
        # Namely, InflectionTemplate should override this.
        raise NotImplementedError

    def add_recording(self, recording):
        pass


class WordformCell(Cell):
    """
    A cell containing a displayable wordform.

    When a ParadigmLayout is filled with forms, the ParadigmLayout.fill() is
    called, converting all its InflectionTemplate instances to WordformCell instances.

    How this differs between **static** and **dynamic** paradigms:
     - **static** paradigms contain ZERO InflectionCell instances; they will all be
        WordformCell instances.
     - **dynamic** paradigms must be filled for every unique FST lemma.
    """

    is_inflection = True

    def __init__(self, inflection: str):
        self.inflection = inflection
        self.recording = None
        self.morphemes = None
        self.add_morphemes()

    def contains_wordform(self, wordform: str) -> bool:
        return self.inflection == wordform

    def add_recording(self, recording):
        self.recording = recording

    def add_morphemes(self):
        analysis = rich_analyze_strict(self.inflection)
        if analysis:
            analysis = analysis[0]
            self.morphemes = analysis.generate_with_morphemes(self.inflection)

    def fill(self, forms: Mapping[str, Collection[str]]) -> tuple[Cell, ...]:
        # No need to fill a cell that already has contents!
        return (self,)

    def __str__(self) -> str:
        return self.inflection

    def __eq__(self, other) -> bool:
        if isinstance(other, WordformCell):
            return self.inflection == other.inflection
        return False

    def __repr__(self):
        name = type(self).__qualname__
        return f"{name}({self.inflection!r})"

    @classmethod
    def parse(cls, text: str):
        assert not looks_like_analysis_string(text)
        return cls(text)


class InflectionTemplate(Cell):
    """
    A cell that contains an inflection.
    """

    is_inflection = True

    def __init__(self, analysis: str):
        assert looks_like_analysis_string(analysis)
        self._analysis_template = string.Template(analysis)

    @property
    def analysis_template(self) -> str:
        """
        The original analysis string provided as a template for inflection,
        e.g., ${lemma}+N+Sg
        """
        return self._analysis_template.template

    def __eq__(self, other) -> bool:
        if isinstance(other, InflectionTemplate):
            return self.analysis_template == other.analysis_template
        return False

    def __str__(self):
        return self.analysis_template

    @staticmethod
    def parse(text: str) -> InflectionTemplate:
        if not looks_like_analysis_string(text):
            raise ParseError(f"cell does not look like an inflection: {text!r}")
        return InflectionTemplate(text)

    def as_analysis(self, lemma: str) -> str:
        """
        Returns a the analysis template with the substitute replaced
        """
        return self._analysis_template.substitute(lemma=lemma)

    def fill(
        self, forms: Mapping[str, Collection[str]]
    ) -> tuple[WordformCell | MissingForm, ...]:
        """
        Return one or more cells with the forms given.
        :raises ParadigmGenerationError: when the analysis is missing from the given
            forms mapping.
        """

        try:
            cell_forms = forms[self.analysis_template]
        except KeyError:
            raise ParadigmGenerationError(
                "no form(s) provided for " f"analysis: {self.analysis_template}"
            )

        if len(cell_forms) == 0:
            # It's a missing form (accidental gap/lacuna).
            # See: https://en.wikipedia.org/wiki/Accidental_gap#Morphological_gaps
            return (MissingForm(),)

        return tuple(WordformCell(form) for form in cell_forms)


class SingletonMixin:
    """
    Mixin that makes any subclasses a singleton.
    """

    def __new__(cls):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
        return cls._instance

    def __repr__(self):
        name = type(self).__qualname__
        return f"{name}()"


class MissingForm(Cell, SingletonMixin):
    """
    A missing form from the paradigm that should show up in the paradigm as a
    placeholder. Missing forms can either be hard-coded or the generator may simply
    not be able to generate a form for a given analysis.

    Note: a missing form is a valid position in the paradigm, however, we display that
    this form cannot exist. This is not the same as an empty cell, which is used as a
    spacer.
    """

    is_inflection = True
    is_missing = True

    def __str__(self):
        return "--"

    def contains_wordform(self, wordform: str) -> bool:
        """
        A missing form does not contain a wordform, be definition.
        """
        return False

    def fill(self, forms: Mapping[str, Collection[str]]) -> tuple[Cell, ...]:
        """
        A missing form is already "filled" -- return one entry
        """
        return (self,)


class EmptyCell(Cell, SingletonMixin):
    """
    A completely empty cell. This is used for spacing in the paradigm. There is no
    semantic content. Compare with MissingForm.
    """

    is_empty = True

    def __str__(self):
        return ""


class SuppressOutputCell(Cell, SingletonMixin):
    """
    Indicates that this should not be rendered at all in the paradigm.
    Note: even an EmptyCell is "output", in that a <tr> should be created for it in HTML.
    A SuppressOutputCell should not get even get a <tr>.
    """

    should_suppress_output = True


class BaseLabelCell(Cell):
    is_label = True
    label_for: str
    prefix: str

    def __init__(self, tags):
        self._tags = tuple(tags)

    @property
    def fst_tags(self) -> tuple[str]:
        return self._tags

    def __str__(self):
        return " ".join(f"{self.prefix} {tag}" for tag in self._tags)

    def __eq__(self, other) -> bool:
        if isinstance(other, type(self)):
            return self._tags == other._tags
        return False

    def __repr__(self):
        name = type(self).__qualname__
        return f"{name}({self._tags!r})"

    @classmethod
    def parse(cls, text: str):
        return cls(parse_label(text, prefix=cls.prefix))


class RowLabel(BaseLabelCell):
    """
    Labels for the cells in the current row within the pane.
    """

    label_for = "row"
    prefix = "_"

    def __init__(self, tags, row_span: int = 1):
        super().__init__(tags)
        self.row_span = row_span

    def with_row_span(self, span: int) -> RowLabel:
        """
        Create a brand new row label with its row span. This is intended so that an
        existing label can be used to span multiple rows.
        """
        return RowLabel(self.fst_tags, span)


class ColumnLabel(BaseLabelCell):
    """
    Labels for the cells in the current column within the pane.
    """

    label_for = "col"
    prefix = "|"


def parse_label(text: str, *, prefix: str) -> list[str]:
    """
    Parses labels in the common label syntax.

    >>> parse_label("| Ind", prefix="|")
    ['Ind']
    >>> parse_label("# Fut # Def", prefix="#")
    ['Fut', 'Def']
    >>> parse_label("_ 1Sg _ 3Sg _ Obv", prefix="_")
    ['1Sg', '3Sg', 'Obv']
    """
    splits = re.split(r" +", text)
    if len(splits) % 2 != 0:
        raise ParseError(f"Uneven number of space-separated segments in {text!r}")

    tags = []
    for actual_prefix, tag in pairs(splits):
        if actual_prefix != prefix:
            raise ParseError(f"Expected prefix {prefix!r} but saw {actual_prefix!r}")
        tags.append(tag)

    return tags


def looks_like_analysis_string(text: str) -> bool:
    """
    Returns true if the cell might be analysis.
    """
    return "${lemma}" in text


def pairs(seq):
    """
    Returns pairs from the given sequence.

    >>> list(pairs([1, 2, 3, 4, 5, 6]))
    [(1, 2), (3, 4), (5, 6)]
    """
    return zip(seq[::2], seq[1::2])


def n_empty_lists(n: int) -> list[list[Cell]]:
    """
    Returns the given number of distinct, empty lists.
    """
    return [[] for _ in range(n)]


def cell_if_exists_or_no_output(col: Sequence[Cell], row_num: int):
    """
    Extracts the content cell from the column if it exists, else returns an EmptyCell if
    nothing is found.
    """
    try:
        return col[row_num]
    except IndexError:
        return SuppressOutputCell()


def adjust_row_span(cell: Cell, span: int) -> Cell:
    """
    Fixes RowLabels such that they span the amount given. Only affects RowLabels.
    Other cells are returned as is.
    """
    if isinstance(cell, RowLabel):
        return cell.with_row_span(span)
    return cell

def rich_analyze_strict(text):
    return list(
        RichAnalysis(r) for r in TransducerFile(STRICT_ANALYSER_FST_FILEPATH).lookup_lemma_with_affixes(text)
    )


class RichAnalysis:
    """The one true FST analysis class.

    Put all your methods for dealing with things like `PV/e+nipâw+V+AI+Cnj+3Pl`
    here.
    """

    def __init__(self, analysis):
        if isinstance(analysis, Analysis):
            self._tuple = analysis
        elif (isinstance(analysis, list) or isinstance(analysis, tuple)) and len(
            analysis
        ) == 3:
            prefix_tags, lemma, suffix_tags = analysis
            self._tuple = Analysis(
                prefixes=tuple(prefix_tags), lemma=lemma, suffixes=tuple(suffix_tags)
            )
        else:
            raise Exception(f"Unsupported argument: {analysis=!r}")

    @property
    def tuple(self):
        return self._tuple

    @property
    def lemma(self):
        return self._tuple.lemma

    @property
    def prefix_tags(self):
        return self._tuple.prefixes

    @property
    def suffix_tags(self):
        return self._tuple.suffixes

    def generate(self):
        return STRICT_GENERATOR_FST_FILEPATH().lookup(self.smushed())

    def generate_with_morphemes(self, inflection):
        try:
            results = TransducerFile(STRICT_GENERATOR_FST_WITH_MORPHEMES_FILEPATH).lookup(self.smushed())
            if len(results) != 1:
                for result in results:
                    if "".join(re.split(r"[<>]", result)) == inflection:
                        return re.split(r"[<>]", result)
                return None
            return re.split(r"[<>]", results[0])
        except RuntimeError as e:
            print("Could not generate morphemes:", e)
            return []

    def smushed(self):
        return "".join(self.prefix_tags) + self.lemma + "".join(self.suffix_tags)

    def tag_set(self):
        return set(self.suffix_tags + self.prefix_tags)

    def tag_intersection_count(self, other):
        """How many tags does this analysis have in common with another?"""
        if not isinstance(other, RichAnalysis):
            raise Exception(f"Unsupported argument: {other=!r}")
        return len(self.tag_set().intersection(other.tag_set()))

    def __iter__(self):
        """Allows doing `head, _, tail = rich_analysis`"""
        return iter(self._tuple)

    def __hash__(self):
        return hash(self._tuple)

    def __eq__(self, other):
        if not isinstance(other, RichAnalysis):
            return NotImplemented
        return self._tuple == other.tuple

    def __repr__(self):
        return f"RichAnalysis({[self.prefix_tags, self.lemma, self.suffix_tags]!r})"



"""
Handles paradigm generation.
"""

from pathlib import Path
from .manager import *
from . import settings

from hfst_optimized_lookup import TransducerFile


def default_paradigm_manager() -> ParadigmManager:
    """
    Returns the ParadigmManager instance that loads layouts and FST from the res
    (resource) directory for the crk/eng language pair (itwêwina).

    Affected by:
      - MORPHODICT_PARADIGM_SIZE_ORDER
    """

    # Directory with paradigm layouts to load
    layout_dir = Path(settings.get_layouts_dir())

    generator = strict_generator()

    if hasattr(settings, "MORPHODICT_PARADIGM_SIZES"):
        return ParadigmManagerWithExplicitSizes(
            layout_dir,
            generator,
            ordered_sizes=settings.MORPHODICT_PARADIGM_SIZES,
        )
    else:
        return ParadigmManager(layout_dir, generator)


def strict_generator():
    """
    FST file to do translations
    """
    path = settings.get_fst_filepath()
    try:
        return TransducerFile(path)
    except Exception as error:
        if str(error).split(":")[0] == 'Transducer not found':
            raise Exception(str(error).split(":")[0] + f' in \"{path}\"')
        else:
            raise Exception(str(error))



from typing import Dict

from .panes import Paradigm
from . import settings


def serialize_paradigm(paradigm: Paradigm) -> Dict[str, any]:
    """
    Serializes a Paradigm object as a dictionary

    :param paradigm: the paradigm to be serialized
    :return: dictionary representation
    """
    panes = []

    if paradigm is None:
        return {"panes": panes}

    for pane in (paradigm.panes or []):

        tr_rows = []
        for row in (pane.rows or []):
            if row.is_header:
                tr_rows.append({
                    "is_header": True,
                    "label": row.fst_tags,
                    "cells": []
                })
            else:
                cells = []
                for cell in (row.cells or []):
                    cell_data = {
                        "should_suppress_output": cell.should_suppress_output,
                        "is_label": cell.is_label,
                        "is_inflection": cell.is_inflection,
                        "is_missing": cell.is_missing,
                        "is_empty": cell.is_empty
                    }

                    if cell_data["is_label"]:
                        cell_data["label_for"] = cell.label_for
                        cell_data["label"] = cell.fst_tags

                        if type(cell_data["label_for"]) != str:  # if cell.label_for was never instantiated
                            cell_data["label_for"] = ""
                        if cell_data["label_for"] == "row":
                            cell_data["row_span"] = cell.row_span

                    elif cell_data["is_inflection"] and not cell_data["is_missing"]:
                        cell_data["inflection"] = cell.inflection

                    cells.append(cell_data)

                tr_rows.append({
                    "is_header": False,
                    "label": None,  # shouldn't be used.
                    "cells": cells
                })

        panes.append({"tr_rows": tr_rows})

    return {"panes": panes}



"""
    Planning to make it a central place to run paradigm generation
"""

from typing import (Collection, Dict, Optional)

from .generation import default_paradigm_manager, ParadigmManager
from .helpers import serialize_paradigm
from .panes import Paradigm
from . import settings


class PaneGenerator:
    def generate_pane(self, lemma: str, paradigm_type: str, specified_size: Optional[str] = None) -> Dict[str, any]:
        """
        Validate given parameters and set options (settings) and return serialized pane of the paradigm"""
        if settings.is_setup_complete():
            if paradigm_type is not None:
                paradigm = self._create_paradigm(lemma, paradigm_type, specified_size)

                serialized_paradigm = serialize_paradigm(paradigm)

                return serialized_paradigm
            else:
                raise Exception("Paradigm layout specification is missing.")
        else:
            raise Exception("FST and Layouts resources are not configured correctly.")

    def _create_paradigm(self, lemma: str, paradigm_type: str, specified_size: Optional[str] = None) -> Paradigm:
        """
        Create the paradigm as a built-in type and return.
        Check given optional size, and use default otherwise.
        """
        paradigm_manager = default_paradigm_manager()
        sizes = list(paradigm_manager.sizes_of(paradigm_type))
        if "basic" in sizes:
            default_size = "basic"
        else:
            default_size = sizes[0]

        if len(sizes) <= 1:
            size = default_size
        else:
            size = specified_size
            if size not in sizes:
                size = default_size

        paradigm = self.paradigm_for(paradigm_manager, paradigm_type, lemma, size)

        return paradigm

    def all_analysis_template_tags(self, paradigm_type: str) -> Collection[tuple]:
        """
        Return the set of all analysis templates in layouts of paradigm_name

        If a paradigm has two sizes, one with template `${lemma}+A` and the
        other with both `${lemma}+A` and `X+${lemma}+B`, then this function will
        return {((), ("+A",)), (("X+",), ("+B",)}.

        Note that these analyses are meant to be inputs to a generator FST for
        building a paradigm table, not the results of analyzing some input
        string.
        """
        if settings.is_setup_complete():
            if paradigm_type is not None:
                pg = default_paradigm_manager()
                all_template_tags = pg.all_analysis_template_tags(paradigm_type)
                return all_template_tags
            else:
                raise Exception("Paradigm layout specification is missing.")
        else:
            raise Exception("FST and Layouts resources are not configured correctly.")

    def paradigm_for(self, paradigm_manager: ParadigmManager, paradigm_type: str, fst_lemma: str, paradigm_size: str) -> \
    Optional[Paradigm]:
        """
        Returns a paradigm for the given wordform at the desired size.

        If a paradigm cannot be found, None is returned
        """

        if paradigm_type:
            if paradigm := paradigm_manager.paradigm_for(paradigm_type, fst_lemma, paradigm_size):
                return paradigm

        return None

    @staticmethod
    def set_fst_filepath(path: str) -> None:
        """
        Set fst resource location
        """
        settings.set_fst_filepath(path)

    @staticmethod
    def set_morpheme_fst_filepath(path: str) -> None:
        """
        Set fst resource location
        """
        settings.set_morpheme_fst_filepath(path)

    @staticmethod
    def set_analyser_fst_filepath(path: str) -> None:
        """
        Set fst resource location
        """
        settings.set_analyser_fst_filepath(path)

    @staticmethod
    def set_layouts_dir(path: str) -> None:
        """
        Set layouts resources location
        """
        settings.set_layouts_dir(path)

    @staticmethod
    def set_tag_style(style: str) -> None:
        """
        Set tag for all templates rendering
        """
        settings.set_tag_style(style)




from pathlib import Path

# The order in which paradigm sizes will be presented to the user.
# The first size in this list is the "default".
# Make sure to exhaustively specify all size options available!
MORPHODICT_PARADIGM_SIZES = [
    # The most user-friendly size should be first:
    "basic",
    # Then, a more complete paradigm layout:
    "full",
    # Variants for linguists go here:
]

STRICT_GENERATOR_FST_FILEPATH = ""
STRICT_GENERATOR_FST_WITH_MORPHEMES_FILEPATH = ""
STRICT_ANALYSER_FST_FILEPATH = ""
LAYOUTS_DIR = ""
TAG_STYLE = "Plus"


class FileDoesNotMatch(Exception):
    """
    Raised when file cannot be found or does not match .hfstol extension
    """


def set_fst_filepath(filepath: str) -> None:
    """
    Set a path to FST file
    """
    if Path(filepath).is_file() and Path(filepath).match("*.hfstol"):
        global STRICT_GENERATOR_FST_FILEPATH
        STRICT_GENERATOR_FST_FILEPATH = filepath
    else:
        raise FileDoesNotMatch(f"file {Path(filepath).as_posix()!r} does not exist or does not end in \".hfstol\".")


def set_morpheme_fst_filepath(filepath: str) -> None:
    """
        Set a path to the morpheme generation FST file
        """
    if Path(filepath).is_file() and Path(filepath).match("*.hfstol"):
        global STRICT_GENERATOR_FST_WITH_MORPHEMES_FILEPATH
        STRICT_GENERATOR_FST_WITH_MORPHEMES_FILEPATH = filepath
    else:
        raise FileDoesNotMatch(f"file {Path(filepath).as_posix()!r} does not exist or does not end in \".hfstol\".")


def set_analyser_fst_filepath(filepath: str) -> None:
    """
        Set a path to the morpheme generation FST file
        """
    if Path(filepath).is_file() and Path(filepath).match("*.hfstol"):
        global STRICT_ANALYSER_FST_FILEPATH
        STRICT_ANALYSER_FST_FILEPATH = filepath
    else:
        raise FileDoesNotMatch(f"file {Path(filepath).as_posix()!r} does not exist or does not end in \".hfstol\".")


def get_fst_filepath() -> str:
    """
    Get a path to FST file
    """
    return STRICT_GENERATOR_FST_FILEPATH


class DirectoryDoesNotExist(Exception):
    """
    Raised when directory with paradigm layouts is missing
    """


def set_layouts_dir(dir_path: str) -> None:
    """
    Set a directory with paradigm layouts
    """
    if Path(dir_path).is_dir():
        global LAYOUTS_DIR
        LAYOUTS_DIR = dir_path
    else:
        raise DirectoryDoesNotExist(f"Directory {dir_path!r} does not exist.")


def get_layouts_dir() -> str:
    """
    Get a directory with paradigm layouts
    """
    return LAYOUTS_DIR


def is_setup_complete():
    """
    Confirms that both FST and Layouts folder are configured and valid.
    """
    return Path(LAYOUTS_DIR).is_dir() and (Path(STRICT_GENERATOR_FST_FILEPATH).is_file() and Path(STRICT_GENERATOR_FST_FILEPATH).match("*.hfstol"))


# In case all_analysis_template_tags() will be need to be called from panes package
def set_tag_style(tag_style: str) -> None:
    """
    Set tag style

    Available tags:
        1. "Plus"
        2. "Bracket"
    """
    global TAG_STYLE
    TAG_STYLE = tag_style


def get_tag_style() -> str:
    """
    Get tag style
    """
    return TAG_STYLE



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['paradigm_panes']

package_data = \
{'': ['*']}

install_requires = \
['hfst-optimized-lookup>=0.0.13,<0.1.0',
 'more-itertools>=8.7.0,<8.8.0',
 'typing-extensions>=3.7,<4.0']

setup_kwargs = {
    'name': 'paradigm-panes',
    'version': '0.3.4',
    'description': 'Paradigm panes meant to provide layout specification to be reused elsewhere.',
    'long_description': '# paradigm-panes\n\nInstallable package that produces a paradigm for a given word, given a pointer to paradigm layouts and FST file. Originally\nbuilt for [itwêwina](https://itwewina.altlab.app/).\n\n# PyPi Package\n\nLatest version of the package posted to PyPi: [paradigm-panes 0.3.2](https://pypi.org/project/paradigm-panes/)\n\n# Install\n\n```\npip install paradigm-panes\n```\n\n# Developing\n\nDeveloping is done and managed through [Python Poetry](https://python-poetry.org/) package manager.\n\nTo start development:\n\n```\n# Download the repo\ngit clone https://github.com/UAlbertaALTLab/paradigm-panes.git\n\n# Set up virutal env\nvirtualenv venv --python=python3.9\nsource venv/bin/activate\n\n# Install dependencies\npoetry install\n\n# Now cd into main directory and try out the package\ncd paradigm_panes\npython\n    >>> import paradigm_panes\n    >>> ...\n```\n\n# API Documentation:\n\n- ## PaneGenerator()\n\n  For successful execution, the package needs a link to resources as described below. To manage different links and use the panes generator more effectively, the package provides PaneGenerator class that allows managing settings and executing main functionality:\n\n  ```\n  >>> pane_generator = paradigm_panes.PaneGenerator()\n  ```\n\n  The class itself does not take any variables.\n\n- ## generate_pane()\n\n  ```\n  >>> paradigm = pane_generator.generate_pane(lemma, paradigm_type, size: Optional)\n  ```\n\n  This function is a core functionality of the package. Once the [resources are specified](#settings-specification-functions), this function generates the pane according to the specification given. \\\n   If the translations are not found in the FST file, some of the inflections will be indicated as missing. The resulted paradigm class is serialized and returned as JSON.\n\n  ### Parameters:\n\n  > lemma(str) - base wordform to be inflected\n\n  > paradigm_type(str) - specification of the paradigm type of the word. Ex. "NA".\n\n  > size(str) - optional size of the pane to be returned. Currently supports "base", and "full". If the specified size is not found, overrides with default option.\n\n- ## all_analysis_template_tags()\n\n  ```\n  >>> paradigm = pane_generator.all_analysis_template_tags(paradigm_type)\n  ```\n\n  An additional functionality that returns all analysis template tags.\n\n  ### Parameters:\n\n  > paradigm_type(str) - specification of the paradigm type of the word. Ex. "NA".\n\n  > tag_style(str) - style of tags to return.\n\n  Specify tag style through [settings](#settings-specification-functions)\n\n# Usage and Configuration\n\nImport the library:\n\n```\n>>> import paradigm_panes\n```\n\nCreate PaneGenerator and specify path to FST file and layouts resources:\n\n```\n>>> pg = paradigm_panes.PaneGenerator()\n>>> pg.set_layouts_dir("/home/ubuntu/paradigm_panes/resources/layouts")\n>>> pg.set_fst_filepath("/home/ubuntu/paradigm_panes/resources/fst/crk-strict-generator.hfstol")\n```\n\nPass lemma, paradigm type, and optional size to generate a pane:\n\n```\n>>> lemma = "amisk"\n>>> p_type = "NA"\n>>> paradigm = pg.generate_pane(lemma, p_type)\n\n>>> p_size = "full"\n>>> full_paradigm = pg.generate_pane(lemma, p_type, p_size)\n```\n\n## Settings specification functions:\n\n- `set_layouts_dir(path)` specifies a location of a directory with paradigm layouts that are relevant for current paradigm generation.\n\n- `set_fst_filepath(path)` specifies FST file location with layout translation that are relevant for current paradigm generation.\n\n- `set_tag_style(path)` specifies template rendering type for all_analysis_template_tags function.\n\n  Available tag styles:\n\n  1.  "Plus"\n  2.  "Bracket"\n\nThe generator must specify both locations (FST, layouts) before generating a paradigm.\n\nSize is optional to paradigm generation; by default a base size (or first available) will be used.\n\n# Testing\n\nTo run the tests you need to install required dependencies, it is easier by using a virtual environment like this:\n\n```\n>>> # Set up virutal env\n>>> virtualenv venv --python=python3.9\n>>> source venv/bin/activate\n>>>\n>>> # Install dependencies\n>>> poetry install\n```\n\nOnce the dependencies are installed you can run tests by calling pytest.\n\n```\n>>> pytest\n```\n\n# Release\n\nPackage version number is sorted in pyproject.toml. With every release to PyPi the version needs to be updated. \\\nBuild the package from the main directory before publishing it:\n\n```\n>>> poetry build\n```\n\nTo publish to Test PyPi use poetry and enter credentials associated with Test PyPi account\n\n```\n>>> poetry publish -r testpypi\n```\n\nTo publish to real PyPi use poetry and enter credentials associated with PyPi\n\n```\n>>> poetry publish\n```\n\nAll relevant package specifications and dependencies are managed in `pyproject.toml` file.\n',
    'author': 'Uladzimir Bondarau',
    'author_email': 'bondarau@ualberta.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/UAlbertaALTLab/paradigm-panes',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import Collection, Iterable, Optional, Protocol

from . import settings
from .helpers import serialize_paradigm

from .panes import Paradigm, ParadigmLayout

# I would *like* a singleton for this, but, currently, it interacts poorly with mypy :/
ONLY_SIZE = "<only-size>"

logger = logging.getLogger(__name__)


class ParadigmDoesNotExistError(Exception):
    """
    Raised when a paradigm is requested, but does not exist.
    """


class ParadigmManager:
    """
    Mediates access to paradigms layouts.

    Loads layouts from the filesystem and can fill the layout with results from a
    (normative/strict) generator FST.
    """

    # Mappings of paradigm name => sizes available => the layout
    _name_to_layout: dict[str, dict[str, ParadigmLayout]]

    def __init__(self, layout_directory: Path, generation_fst: Transducer):
        self._generator = generation_fst
        self._name_to_layout: dict[str, ParadigmLayout] = {}

        self._load_layouts_from(layout_directory)

    def paradigm_for(
        self,
        paradigm_name: str,
        lemma: Optional[str] = None,
        size: Optional[str] = None,
    ) -> Paradigm:
        """
        Returns a paradigm for the given paradigm name. If a lemma is given, this is
        substituted into the dynamic paradigm.

        :raises ParadigmDoesNotExistError: when the paradigm name cannot be found.
        """
        layout_sizes = self._layout_sizes_or_raise(paradigm_name)
        if size is None:
            size = self.default_size(paradigm_name)

        if size not in layout_sizes:
            raise ParadigmDoesNotExistError(f"size {size!r} for {paradigm_name}")
        layout = layout_sizes[size]

        if lemma is not None:
            return self._inflect(layout, lemma)
        else:
            return layout.as_static_paradigm()

    def sizes_of(self, paradigm_name: str) -> Collection[str]:
        """
        Returns the size options of the given paradigm.

        :raises ParadigmDoesNotExistError: when the paradigm name cannot be found.
        """
        return self._layout_sizes_or_raise(paradigm_name).keys()

    def all_analyses(self, paradigm_name: str, lemma: str) -> set[str]:
        """
        Returns all analysis strings for a given paradigm and lemma in all layout sizes.

        For example, in Plains Cree, you want all analyses for mîcisow (VAI):

            {"mîcisow+V+AI+Ind+Prs+1Sg", "mîcisow+V+AI+Ind+Prs+2Sg", ...}

        :raises ParadigmDoesNotExistError: when the paradigm name cannot be found.
        """

        analyses: set[str] = set()
        for layout in serialize_paradigm(self._layout_sizes_or_raise(paradigm_name)).values():
            analyses.update(layout.generate_fst_analyses(lemma).values())

        return analyses

    def default_size(self, paradigm_name: str):
        sizes = list(self.sizes_of(paradigm_name))
        return sizes[0]

    def _layout_sizes_or_raise(self, paradigm_name) -> ParadigmLayout:
        """
        Returns the sizes of the paradigm with the given name.

        :raises ParadigmDoesNotExistError: when the paradigm name cannot be found.
        """
        try:
            return self._name_to_layout[paradigm_name]
        except KeyError:
            raise ParadigmDoesNotExistError(paradigm_name)

    def _load_layouts_from(self, path: Path):
        """
        Loads all .tsv files in the path as paradigm layouts.

        Does nothing if the directory does not exist.
        """
        if not Path(path).exists():
            logger.debug("No layouts found in %s", path)
            return

        for paradigm_name, size, layout in _load_all_layouts_in_directory(path):
            # .setdefault() creates a new, empty dict if the paradigm name does not
            # exist yet:
            self._name_to_layout.setdefault(paradigm_name, {})[size] = layout

    _LITERAL_LEMMA_RE = re.compile(r"\$\{lemma\}")

    def all_analysis_template_tags(self, paradigm_name) -> Collection[tuple]:
        """Return the set of all analysis templates in layouts of paradigm_name

        If a paradigm has two sizes, one with template `${lemma}+A` and the
        other with both `${lemma}+A` and `X+${lemma}+B`, then this function will
        return {((), ("+A",)), (("X+",), ("+B",)}.

        Note that these analyses are meant to be inputs to a generator FST for
        building a paradigm table, not the results of analyzing some input
        string.
        """
        ret = {}
        for layout in serialize_paradigm(self._name_to_layout[paradigm_name]).values():
            # The trick here is that we can look for a literal `${lemma}`
            # instead of having to parse arbitrary FST analyses.
            for template in layout.generate_fst_analyses("${lemma}"):
                prefix, suffix = self._LITERAL_LEMMA_RE.split(template)

                if settings.get_tag_style() == "Plus":
                    prefix_tags = prefix.split("+")
                    assert (
                        prefix_tags[-1] == ""
                    ), f"Prefix {prefix!r} did not end with +"
                    suffix_tags = suffix.split("+")
                    assert suffix_tags[0] == "", f"Suffix {suffix!r} did not end with +"
                    ret[template] = (
                        tuple(t + "+" for t in prefix_tags[:-1]),
                        tuple("+" + t for t in suffix_tags[1:]),
                    )
                elif settings.get_tag_style() == "Bracket":
                    ret[template] = (split_brackets(prefix), split_brackets(suffix))
                else:
                    raise Exception(f"Unsupported {settings.get_tag_style()=!r}")
        return ret.values()

    def _inflect(self, layout: ParadigmLayout, lemma: str) -> Paradigm:
        """
        Given a layout and a lemma, produce a paradigm with forms generated by the FST.
        """
        template2analysis = layout.generate_fst_analyses(lemma=lemma)
        analysis2forms = self._generator.bulk_lookup(list(template2analysis.values()))
        template2forms = {
            template: analysis2forms[analysis]
            for template, analysis in template2analysis.items()
        }
        return layout.fill(template2forms)


_BRACKET_SEPATOR_RE = re.compile(
    r"""
            # regex to match the zero-width pattern in the middle of "]["
            (?<= # look-behind
                \] # literal ]
            )
            (?= # look-ahead
                \[ # literal [
            )
           """,
    re.VERBOSE,
)


def split_brackets(s):
    if s == "":
        return []
    return _BRACKET_SEPATOR_RE.split(s)


class ParadigmManagerWithExplicitSizes(ParadigmManager):
    """
    A ParadigmManager but its sizes are always returned, sorted according the explicit
    order specified.
    """

    def __init__(
        self,
        layout_directory: Path,
        generation_fst: Transducer,
        *,
        ordered_sizes: list[str],
    ):
        super().__init__(layout_directory, generation_fst)
        self._size_to_order = {
            element: index for index, element in enumerate(ordered_sizes)
        }

    def sizes_of(self, paradigm_name: str) -> Collection[str]:
        unsorted_results = super().sizes_of(paradigm_name)
        if len(unsorted_results) <= 1:
            return unsorted_results
        return sorted(unsorted_results, key=self._sort_by_explict_order)

    def _sort_by_explict_order(self, element: str) -> int:
        """
        Orders elements according to the given ordered sizes.
        Can be used as a key function for sort() or sorted().
        """
        return self._size_to_order[element]

    def all_sizes_fully_specified(self):
        """
        Returns True when all size options for all paradigms are specified in the
        explicit order given in the constructor.
        """
        valid_sizes = {ONLY_SIZE} | self._size_to_order.keys()
        all_paradigms = self._name_to_layout.keys()

        for paradigm in all_paradigms:
            # use super() to avoid any ordering stuff.
            sizes_available = super().sizes_of(paradigm)
            for size in sizes_available:
                if size not in valid_sizes:
                    logger.error(
                        "Paradigm %r has a layout in size %r, however that "
                        "size has not been declared",
                        paradigm,
                        size,
                    )
                    return False

        return True


def _load_all_layouts_in_directory(path: Path):
    """
    Yields (paradigm, size, layout) tuples from the given directory. Immediate
    subdirectories are assumed to be paradigms with multiple size options.
    """
    assert Path(path).is_dir()

    for filename in Path(path).iterdir():
        if Path(filename).is_dir():
            yield from _load_all_sizes_for_paradigm(filename)
        elif Path(filename).match("*.tsv"):
            yield Path(filename).stem, ONLY_SIZE, _load_layout_file(filename)


def _load_all_sizes_for_paradigm(directory: Path):
    """
    Yields (paradigm, size, layout) tuples for ONE paradigm name. The paradigm name
    is inferred from the directory name.
    """
    paradigm_name = directory.name
    assert directory.is_dir()

    for layout_file in directory.glob("*.tsv"):
        size = layout_file.stem
        assert size != ONLY_SIZE, f"size name cannot clash with sentinel value: {size}"
        yield paradigm_name, size, _load_layout_file(layout_file)


def _load_layout_file(layout_file: Path):
    return ParadigmLayout.loads(Path(layout_file).read_text(encoding="UTF-8"))


class Transducer(Protocol):
    """
    Interface for something that can lookup forms in bulk.

    This is basically the subset of the hfst_optimized_lookup.TransducerFile API that
    the paradigm manager actually uses.
    """

    def bulk_lookup(self, strings: Iterable[str]) -> dict[str, set[str]]:
        ...
